"""Functions for working with coincidence maps.

"""
import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np



def plot_coinc_map(coincmap, xedges, yedges, figsize=None, cmap='YlOrRd',
                   title=None, norm=None, vmin=1, vmax=None, num=None,
                   xlabel="early electron kinetic energy",
                   ylabel="late electron kinetic energy"):
    """Plot a coincedence map and its projections on the x and y axes.

    Parameters
    ----------
    coincmap : numpy.ndarray
        2d array of shape (m,n) containing the coincidence map. In most
        cases this will be the output of ``numpy.histogram2d()``.
    xedges : numpy.ndarray
        1d array of shape (m+1) containing the bin edges of the x-axis.
    yedges : numpy.ndarray
        1d array of shape (n+1) containing the bin edges of the y-axis.
    figsize : tuple, optional
        Figure size in inches. Default: None
    cmap : matplotlib.colors.Colormap or str or None, optional
        Colormap passed to ``matplotlib.pyplot.pcolormesh()``.
        Default: 'YlOrRd'
    title : str, optional
        Title of the figure. Default: None
    norm : str or matplotlib.colors.Normalize or None, optional
        Normalization passed to ``matplotlib.pyplot.pcolormesh()``.
        Default: None
    vmin, vmax : float, optional
        Minimum and maximum value for the colormap passed to
        ``matplotlib.pyplot.pcolormesh()``. Default: 1, None
    num: int or str or matplotlib.figure.Figure, optional
        Figure identifier passed to ``matplotlib.pyplot.figure()``.
    xlabel, ylabel : str, optional
        Labels of the x and y axes. Default:
        "early electron kinetic energy", "late electron kinetic energy"

    Returns
    -------
    fig : matplotlib.figure.Figure
        Matplotlib Figure object.
    ax: tuple of matplotlib.axes.Axes
        Tuple of matplotlib Axes objects containing the coincidence map,
        the projection on the x-axis and the projection on the y-axis.

    """
    fig = plt.figure(num=num, figsize=figsize, clear=True)

    # grid with columns=2, row=2
    gs = gridspec.GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[1, 3],
                           wspace=0.05, hspace=0.05)
    # coinc matrix is subplot 2: lower left
    ax_coinc = plt.subplot(gs[2])
    # spectrum of E0 is subplot 0: upper left
    ax_x = plt.subplot(gs[0], sharex=ax_coinc)
    # spectrum of E1 is subplot 3: lower right
    ax_y = plt.subplot(gs[3], sharey=ax_coinc)
    # colorbar is subplot 1: upper right
    ax_cb = plt.subplot(gs[1])

    # sum spectrum of E0 (top)
    hist_x = np.sum(coincmap, axis=1)
    ax_x.set_xlim(xedges[0], xedges[-1])
    ax_x.stairs(hist_x, xedges, color='k')

    # sum spectrum of E1 (right)
    hist_y = np.sum(coincmap, axis=0)
    ax_y.set_ylim(yedges[0], yedges[-1])
    ax_y.stairs(hist_y, yedges, color='k', orientation="horizontal")

    # coinc matrix
    X, Y = np.meshgrid(xedges, yedges)
    pcm = ax_coinc.pcolormesh(X, Y, coincmap.T, cmap=cmap, norm=norm,
                              vmin=vmin, vmax=vmax, rasterized=True)

    # Generate a colorbar for the histogram in the upper right panel
    ax_cb.axis("off")
    ax_cb_inset = ax_cb.inset_axes([0.0, 0.0, 0.25, 1.0])
    fig.colorbar(pcm, cax=ax_cb_inset)

    # Set the labels
    ax_coinc.set_xlabel(xlabel)
    ax_coinc.set_ylabel(ylabel)

    # Remove x and y tick labels
    ax_x.tick_params(axis='both', labelleft=False, labelbottom=False)
    ax_y.tick_params(axis='both', labelleft=False, labelbottom=False)

    # Set the title
    if title is not None:
        fig.suptitle(title)

    return fig, (ax_coinc, ax_x, ax_y)



"""Functions for working with hysteresis measurement data.

Currently only includes Faraday correction via analytic optimization.
Faraday correction based on measured data can be added on request.

"""

from typing import Sequence
from numpy.typing import NDArray, ArrayLike
import numpy as np
from scipy.optimize import curve_fit, minimize, minimize_scalar
import logging

logger = logging.getLogger(__name__)


def normalize(*magnetization: Sequence,
              saturation_mean_percentage: float = 10) -> NDArray:
    """Normalizes hysteresis to the saturation region.

    Magnetization is normalized to the averaged saturation over a given
    range. The range over which to average is computed from a
    percentage relative to the number of data points in one branch.
    The magnetization needs to be supplied either as one continous
    sequence containing both hysteresis branches or both branches as
    separate sequences of uniform length. In the latter case, the
    second branch needs to be sorted by time recorded, not by magnetic
    field value.

    Parameters
    ----------
    *magnetization: Sequence
        One or two sequences containing the hysteresis branches of the
        magnetization.
    saturation_mean_percentage: float, default 10
        Percentage of half the length of ``magnetization`` over which
        to average the saturation value.

    Returns
    -------
    NDArray
        The normalized magnetization values. If ``magnetization`` is
        one continous array, then a 1-D array with the same length is
        returned. If ``magnetization`` is two arrays, a 2-D array is
        returned containing the two branches in the first axis.

    """

    if len(magnetization) == 1:
        branch_1, branch_2 = np.split(np.array(magnetization[0]), 2)

    if len(magnetization) == 2:
        branch_1, branch_2 = np.array(magnetization)

    if len(magnetization) > 2:
        raise ValueError("Too many input sequences.")

    # calculate number of points over wich to average saturation
    # ceil ensures the presence of at least one data point
    mean_range = np.ceil(len(branch_1) * saturation_mean_percentage
                         / 100).astype(int)

    # normalize
    max_mag = np.mean(np.concatenate((branch_1[:mean_range],
                                      branch_2[len(branch_2) - mean_range:])))
    min_mag = np.mean(np.concatenate((branch_2[:mean_range],
                                      branch_1[len(branch_2) - mean_range:])))

    branch_1 = 2 * (branch_1 - min_mag) / (max_mag - min_mag) - 1
    branch_2 = 2 * (branch_2 - min_mag) / (max_mag - min_mag) - 1

    if len(magnetization) == 1:
        return np.concatenate((branch_1, branch_2))

    if len(magnetization) == 2:
        return np.array((branch_1, branch_2))


def shift_branch(*magnetization: Sequence, shift_value: float) -> NDArray:
    """Shifts hysteresis branches relative to each other.

    The shift is applied to the second branch.

    Parameters
    ----------
    *magnetization: Sequence
        One or two sequences containing the hysteresis branches of the
        magnetization.
    shift_value: float
        Value added to the second hysteresis branch.

    Returns
    -------
    NDArray
        The shifted magnetization values. If ``magnetization`` is
        one continous array, then a 1-D array with the same length is
        returned. If ``magnetization`` is two arrays, a 2-D array is
        returned containing the two branches in the first axis.

    """

    if len(magnetization) == 1:
        branch_1, branch_2 = np.split(np.array(magnetization[0]), 2)

    if len(magnetization) == 2:
        branch_1, branch_2 = np.array(magnetization)

    if len(magnetization) > 2:
        raise ValueError("Too many input sequences.")

    # python built-in lists don't support += operator for maths
    branch_2 = np.array(branch_2)
    branch_2 = branch_2 + shift_value

    if len(magnetization) == 1:
        return np.concatenate((branch_1, branch_2))

    if len(magnetization) == 2:
        return np.array((branch_1, branch_2))


def drift_correction(*hysteresis: ArrayLike, slope: float) -> NDArray:
    """Corrects thermal drift.

    Thermal correction is applied by subtracting a linear function from
    the time ordered hysteresis loop.
    Hysteresis data can be supplied either as one tuple containing the
    magntic field and magnetization repectively or as multiple tuples
    containing the two hysteresis branches.
    This function is purely for subtracting a linear function in time
    domain and does not determine the optical slope for the drift
    correction itself.

    Parameters
    ----------
    *hysteresis: ArrayLike
        One or two 2-D arrays of shape (2, n) containing the
        magnetic field and magnetization of length n in the first axis.
    slope: float
        Slope of the linear function ``M(H) = slope * H`` to subtract.

    Returns
    -------
    NDArray
        The corrected magnetization values. If ``hysteresis`` is
        supplied as whole sequence , then a 1-D array with the same
        length as the magnetization is returned. If ``hysteresis`` is
        two branches, a 2-D array is returned containing the
        two magnetization branches in the first axis.


    Examples
    --------
    Viable example inputs.

    >>> drift_correction((H, M), slope=10)
    array([<M-values>])

    >>> drift_correction((H1, M1), (H2, M2), slope=10)
    array([[<M1-values>], [<M2-values>]])

    >>> hyst = array([H, M])
    >>> drift_correction(hyst, slope=10)
    array([<M-values>])

    >>> hyst = array([[H1, M1], [H2, M2]])
    >>> drift_correction(*hyst, slope=10)
    array([[<M1-values>], [<M2-values>]])

    """

    if len(hysteresis) == 1:
        H, M = np.array(hysteresis[0])
        H_1, H_2 = np.split(H, 2)
        M_1, M_2 = np.split(M, 2)

    if len(hysteresis) == 2:
        (H_1, M_1), (H_2, M_2) = np.array(hysteresis)
        # to support math operations for built-in lists

    if len(hysteresis) > 2:
        raise ValueError("Too many input sequences.")

    M_1 = M_1 + H_1 * slope
    offset = M_1[-1] - M_2[0] + H_2[0] * slope  # preserves line continuity
    M_2 = M_2 - H_2 * slope + offset

    if len(hysteresis) == 1:
        return np.concatenate((M_1, M_2))

    if len(hysteresis) == 2:
        return np.array((M_1, M_2))


def linear_correction(*hysteresis: ArrayLike,
                      saturation_percentage: float = 10) -> NDArray:
    """Subtracts a linear fit to the saturation regions.

    Linear fit is performed on the saturation region with respect to the
    magnetic field. The saturation region is defined by a percentage
    value relative to the number of data points of a single branch.
    A function ``f(x) = fit_slope * x`` is then subtracted from the
    magnetization.
    Hysteresis data can be supplied either as one tuple containing the
    magntic field and magnetization repectively or as multiple tuples
    containing the two hysteresis branches.

    Parameters
    ----------
    *hysteresis: ArrayLike
        One or two 2-D arrays of shape (2, n) containing the
        magnetic field and magnetization of length n in the first axis.
    saturation_percentage: float, default 10
        Percentage value of points in at the end of the hysteresis
        branches belonging to the saturation region.

    Returns
    -------
    NDArray
        The corrected magnetization values. If ``hysteresis`` is
        supplied as whole sequence , then a 1-D array with the same
        length as the magnetization is returned. If ``hysteresis`` is
        two branches, a 2-D array is returned containing the
        two magnetization branches in the first axis.

    Examples
    --------
    Viable example inputs.

    >>> linear_correction((H, M), saturation_percentage=10)
    array([<M-values>])

    >>> linear_correction((H1, M1), (H2, M2), saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    >>> hyst = array([H, M])
    >>> linear_correction(hyst, saturation_percentage=10)
    array([<M-values>])

    >>> hyst = array([[H1, M1], [H2, M2]])
    >>> linear_correction(*hyst, saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    """

    if len(hysteresis) == 1:
        H, M = np.array(hysteresis[0])
        H_1, H_2 = np.split(H, 2)
        M_1, M_2 = np.split(M, 2)

    if len(hysteresis) == 2:
        (H_1, M_1), (H_2, M_2) = np.array(hysteresis)
        # to support math operations for built-in lists
        H_1, M_1 = np.array(H_1), np.array(M_1)
        H_2, M_2 = np.array(H_2), np.array(M_2)

    if len(hysteresis) > 2:
        raise ValueError("Too many input sequences.")

    # ceil ensures at least one data point
    sat_range = np.ceil(len(H_1) * saturation_percentage / 100).astype(int)
    H1_sat = np.concatenate((H_1[:sat_range], H_2[len(H_1) - sat_range:]))
    H2_sat = np.concatenate((H_2[:sat_range], H_1[len(H_1) - sat_range:]))
    M1_sat = np.concatenate((M_1[:sat_range], M_2[len(H_1) - sat_range:]))
    M2_sat = np.concatenate((M_2[:sat_range], M_1[len(H_1) - sat_range:]))

    # fit linear function to saturation region only
    fit_para1, _ = curve_fit(lambda x, m, t: m*x+t, H1_sat, M1_sat)
    fit_para2, _ = curve_fit(lambda x, m, t: m*x+t, H2_sat, M2_sat)
    fit_para = np.mean((fit_para1, fit_para2), axis=0)

    # subtract slope from entire magnetization
    M_1 = M_1 - fit_para[0] * H_1
    M_2 = M_2 - fit_para[0] * H_2

    if len(hysteresis) == 1:
        return np.concatenate((M_1, M_2))

    if len(hysteresis) == 2:
        return np.array((M_1, M_2))


def faraday_correction(*hysteresis: ArrayLike,
                       initial_guess: Sequence[float],
                       saturation_percentage: float = 10,
                       log_level: int = logging.INFO) -> NDArray:
    """Performs a correction of the Faraday-effect via analytic
    optimization method.

    Note: It is recommended to use ``optimzed_verdet_correction``
    instead as it removes the additional guess of the cos amplitude.

    Performs optimization of a functional describing the difference
    between saturation region and analytic expression of the Faraday-
    effect. The correction might need to be combined with other
    correction methods (thermal drift, relative shifting, ...) to obtain
    proper looking hysteresis curves.
    Note, that the optimized functional considers the linearly
    corrected data, but the function itself does not subtract the linear
    part.
    Hysteresis data can be supplied either as one tuple containing the
    magntic field and magnetization repectively or as multiple tuples
    containing the two hysteresis branches.

    Parameters
    ----------
    *hysteresis: ArrayLike
        One or two 2-D arrays of shape (2, n) containing the
        magnetic field and magnetization of length n in the first axis.
    initial_guess: Sequence[float, float, float]
        Initial guess for verdet constant of the objective lens,
        amplitude and left shift of cos^2 function.
    saturation_percentage: float, default 10
        Percentage value of points in at the end of the hysteresis
        branches belonging to the saturation region.
    log_level: int, default 10
        log behavior of ``python.logging`` module for the optimization
        results.

    Returns
    -------
    NDArray
        The corrected magnetization values. If ``hysteresis`` is
        supplied as whole sequence , then a 1-D array with the same
        length as the magnetization is returned. If ``hysteresis`` is
        two branches, a 2-D array is returned containing the
        two magnetization branches in the first axis.

    Examples
    --------
    Viable example inputs.

    >>> faraday_correction((H, M), initial_guess=(0.01, 5, 100),
                            saturation_percentage=10)
    array([<M-values>])

    >>> faraday_correction((H1, M1), (H2, M2), initial_guess=(0.01, 5, 100),
                            saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    >>> hyst = array([H, M])
    >>> faraday_correction(hyst, initial_guess=(0.01, 5, 100),
                            saturation_percentage=10)
    array([<M-values>])

    >>> hyst = array([[H1, M1], [H2, M2]])
    >>> faraday_correction(*hyst, initial_guess=(0.01, 5, 100),
                            saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    """

    if len(hysteresis) == 1:
        H, M = np.array(hysteresis[0])

    if len(hysteresis) == 2:
        (H_1, M_1), (H_2, M_2) = np.array(hysteresis)
        H = np.concatenate((H_1, H_2))
        M = np.concatenate((M_1, M_2))

    if len(hysteresis) > 2:
        raise ValueError("Too many input sequences.")

    # ceil ensures at least one data point
    sat_range = np.ceil(len(H) / 2 * saturation_percentage / 100).astype(int)

    def cos_sq(x, A, v, s):
        """Parameterized cos^2 function.

        Parameters
        ----------
        x: float
            function input.
        A: float
            amplitude
        v: float
            verdet constant
        s: float
            left shift

        """

        return A * np.cos(v * (x + s) + np.pi/2)**2

    def functional(x):
        """Functional to optimize.

        Assumes the ideal correction is reached, when the saturation
        flattens out.

        """

        cos_squared = cos_sq(H, *x)
        corrected = linear_correction(
            (H, M - cos_squared), saturation_percentage=saturation_percentage)
        derivative = np.gradient(corrected)
        branch_1, branch_2 = np.split(derivative, 2)
        saturation = np.concatenate((
            branch_1[:sat_range], branch_1[len(branch_1) - sat_range:],
            branch_2[:sat_range], branch_2[len(branch_2) - sat_range:]))

        return np.mean(np.abs(saturation))

    # optimize the functional
    minimum = minimize(functional, x0=initial_guess)
    logger.log(log_level,
               f"Computed minimum for faraday correction:\n{minimum}")
    M_corrected = M - cos_sq(H, *(minimum.x))

    if len(hysteresis) == 1:
        return M_corrected

    if len(hysteresis) == 2:
        return np.array(np.split(M_corrected, 2))


def optimized_faraday_correction(*hysteresis: ArrayLike,
                                 verdet_guess: float,
                                 left_shift_guess: float,
                                 bracket: Sequence[float] = None,
                                 bounds: Sequence[float] = None,
                                 saturation_percentage: float = 10,
                                 log_level: int = logging.INFO) -> NDArray:
    """Performs a correction of the Faraday-effect via analytic
    optimization method.

    Performs optimization of a functional describing the difference
    between saturation region and analytic expression of the Faraday-
    effect. The correction might need to be combined with other
    correction methods (thermal drift, relative shifting, ...) to obtain
    proper looking hysteresis curves.
    Note, that the optimized functional considers the linearly
    corrected data, but the function itself does not subtract the linear
    part.
    Hysteresis data can be supplied either as one tuple containing the
    magntic field and magnetization repectively or as multiple tuples
    containing the two hysteresis branches.

    Parameters
    ----------
    *hysteresis: ArrayLike
        One or two 2-D arrays of shape (2, n) containing the
        magnetic field and magnetization of length n in the first axis.
    verdet_guess: float
        Initial guess for verdet constant of the objective lens.
    left_shift_guess: float
        Initial guess for the extremum position of the cos^2 function.
    bracket: Sequence[float, float(, float)], default None
        Bracket parameter of ``scipy.optimize.minimize_scalar``.
        Defines the bracketing interval.
        Must be supplied if ``bounds`` is ``None``.
    bounds: Sequence[float, float]
        Bounds parameter of ``scipy.optimizle.minimize_scalar``.
        Defines the upper and lower bounds of the guess values.
        Must be supplied if ``bracket`` is ``None``.
    saturation_percentage: float, default 10
        Percentage value of points in at the end of the hysteresis
        branches belonging to the saturation region.
    log_level: int, default 10
        log behavior of ``python.logging`` module for the optimization
        results.

    Returns
    -------
    NDArray
        The corrected magnetization values. If ``hysteresis`` is
        supplied as whole sequence , then a 1-D array with the same
        length as the magnetization is returned. If ``hysteresis`` is
        two branches, a 2-D array is returned containing the
        two magnetization branches in the first axis.

    Examples
    --------
    The optimization parameters for this function try to describe the
    curvature in a hysteresis curve with a cos^2 function. The
    verdet constant is a material constant, that in this cas scales the
    width of the cos^2 function. For usual field ranges (10 - 100 mT)
    it is quite small and is in the proximity of 0.01 or lower.
    The left shift describes were the extremum of the of the cos^2 is
    and is in field units. This parameter can usually be estimated by
    plotting the hysteresis beforehand. The ``bounds`` or ``bracket``
    parameters describe the amplitude of the cos^2 and can take values
    from the range of the distance between saturation branches up until
    several times of that.

    The following example demonstrates a typical correction code.

    >>> M, H = ...
    >>> M = drift_correction((H, M), slope=0.2)
    >>> M = shift_branch(M, shift_value=5)
    >>> M = optimized_faraday_correction((H, M),
                                         verdet_guess=0.01,
                                         left_shift_guess=5,
                                         bounds=(20, 500),
                                         saturation_percentage=30)
    >>> M = linear_correction((H, M), saturation_percentage=30)

    The first two functions clean the raw data from potential thermal
    drift and tries to align the two hysteresis branches as good as
    possible. The better the alignment beforehand, the better the
    faraday correction works. It is suggested to plot the intermediate
    states and play with the parameters until it aligns best.

    The faraday correction is then applied with the recommended guess
    parameters. Don't expect these to work right out of the box though.
    Sometimes, one has to play around with the guess parameters to find
    a good solution. For some measurement data it might not work at all
    due to particular strong distortions or other effects. Generally, a
    longer saturation region in the input data yields better results.

    Finally, a linear correction is applied to remove any remaining
    hysteresis tilt. The linear correction is considered in the
    optimized functional of the faraday correction. However, the
    correction itself only removes the faraday contribution itself. For
    a clean hysteresis this last step is therefore necessary.

    Viable example inputs.

    >>> faraday_correction((H, M), verdet_guess=0.01, left_shift_guess=5,
                           saturation_percentage=10)
    array([<M-values>])

    >>> faraday_correction((H1, M1), (H2, M2), verdet_guess=0.01,
                           left_shift_guess=5,
                           saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    >>> hyst = array([H, M])
    >>> faraday_correction(hyst, verdet_guess=0.01, left_shift_guess=5,
                           saturation_percentage=10)
    array([<M-values>])

    >>> hyst = array([[H1, M1], [H2, M2]])
    >>> faraday_correction(*hyst, verdet_guess=0.01, left_shift_guess=5,
                           saturation_percentage=10)
    array([[<M1-values>], [<M2-values>]])

    """

    if len(hysteresis) == 1:
        H, M = np.array(hysteresis[0])

    if len(hysteresis) == 2:
        (H_1, M_1), (H_2, M_2) = np.array(hysteresis)
        H = np.concatenate((H_1, H_2))
        M = np.concatenate((M_1, M_2))

    if len(hysteresis) > 2:
        raise ValueError("Too many input sequences.")

    # ceil ensures at least one data point
    sat_range = np.ceil(len(H)/2 * saturation_percentage / 100).astype(int)

    def functional(scale):
        """Wrapping functional for determining the correct amplitude.

        Assumes the ideal correction is reached, when the saturation
        flattens out.

        """

        corrected = faraday_correction(
            (H, M), initial_guess=(scale, verdet_guess, left_shift_guess),
            saturation_percentage=saturation_percentage,
            log_level=logging.DEBUG
        )
        derivative = np.gradient(corrected)
        branch_1, branch_2 = np.split(derivative, 2)
        saturation = np.concatenate((
            branch_1[:sat_range], branch_1[len(branch_1) - sat_range:],
            branch_2[:sat_range], branch_2[len(branch_2) - sat_range:]))

        return np.mean(np.abs(saturation))

    # optimize the functional
    if bracket is not None:
        minimum = minimize_scalar(functional, bracket=bracket)
    if bounds is not None:
        minimum = minimize_scalar(functional, bounds=bounds, method="bounded")

    logger.log(log_level, f"Computed minimum for amplitude guess:\n{minimum}")
    M_corrected = faraday_correction(
        (H, M), initial_guess=(minimum.x, verdet_guess, left_shift_guess),
        saturation_percentage=saturation_percentage, log_level=log_level
    )

    if len(hysteresis) == 1:
        return M_corrected

    if len(hysteresis) == 2:
        return np.array(np.split(M_corrected, 2))



"""Processing and analysis of fluorescence spectra.

"""
from __future__ import annotations
from typing import Union, Tuple, Sequence
import warnings

import numpy as np
import h5py

from ._anodes import *


class Spectrum:
    """Fluorescence spectrum.

    """

    def __init__(self,
        raw: np.ndarray,
        time: int = None,
        anode: PositionAnode = None,
        target_density: float = None,
        intensity_downstream: float = None,
        intensity_upstream: float = None,
    ) -> None:
        # Store the passed data
        self._raw = raw
        self._t = time
        self.anode = anode
        self._target_density = target_density
        self._intensity_downstream = intensity_downstream
        self._intensity_upstream = intensity_upstream

    @classmethod
    def from_h5(cls,
        file_path: str,
        raw: str = "dld_rd#raw/0/0.0",
        time: int = None,
        anode: PositionAnode = None,
        target_density: str = None,
        intensity_downstream: str = None,
        intensity_upstream: str = None,
    ) -> None:
        with h5py.File(file_path, "r") as h5:
            # Load the raw data
            raw = np.asarray(h5[raw])
            # Load the target density
            if target_density is not None:
                target_density = h5[target_density][0]
                print("Target density:", target_density)
            # Load the upstream intensity
            if intensity_upstream is not None:
                intensity_upstream = h5[intensity_upstream][0]
            # Load the downstream intensity
            if intensity_downstream is not None:
                intensity_downstream = h5[intensity_downstream][0]
        # Initialize the Spectrum
        return cls(raw, time=time, anode=anode, target_density=target_density,
                   intensity_downstream=intensity_downstream,
                   intensity_upstream=intensity_upstream)

    @classmethod
    def from_scan(cls,
        file_path: str,
        scan_var: str = None,
        raw: str = "dld_rd#raw/0",
        time_per_step: int = None,
        anode: PositionAnode = None,
        target_density: str = None,
        intensity_downstream: str = None,
        intensity_upstream: str = None,
    ) -> Tuple[list, list]:
        with h5py.File(file_path, "r") as h5:
            # Load the steps
            if scan_var is not None:
                steps = h5[scan_var]
            # Load the raw data
            raw = h5[raw]
            # Load the target density
            if target_density is not None:
                target_density = np.asarray(h5[target_density])
            else:
                target_density = np.full(len(raw), None)
            # Load upstream intensity
            if intensity_upstream is not None:
                intensity_upstream = np.asarray(h5[intensity_upstream])
            else:
                intensity_upstream = np.full(len(raw), None)
            # Load downstream intensity
            if intensity_downstream is not None:
                intensity_downstream = np.asarray(h5[intensity_downstream])
            else:
                intensity_downstream = np.full(len(raw), None)
            # Format the data and steps
            spectra = []
            step_val = []
            for i, step in enumerate(raw.keys()):
                # Format the step value
                if scan_var is not None:
                    step_val.append(steps[step][0][0])
                else:
                    step_val.append(float(step))
                # Format the raw data
                data = np.asarray(raw[step])
                # Initialize the spectrum instance
                spectra.append(cls(
                    data,
                    time=time_per_step,
                    anode=anode,
                    target_density=target_density[i],
                    intensity_downstream=intensity_downstream[i],
                    intensity_upstream=intensity_upstream[i]
                ))
        # Return the spectra and energies
        return spectra, step_val

    def counts(self,
        anode: PositionAnode = None,
        roi: dict = None,
        efficiency_map: Tuple[np.ndarray, np.ndarray, np.ndarray] = None,
        background: Spectrum = None,
    ) -> Tuple[float, float]:
        """Get number of counts in the spectrum and the estimated
        uncertainty.

        Parameters
        ----------
        anode: PositionAnode
            Anode object from `agepy.spec.photons`.
        """
        # Get the xy values
        if anode is None:
            anode = self.anode
        if anode is None:
            raise ValueError("Anode object must be provided for processing.")
        det_image = anode.process(self._raw)
        # Use the full detector if roi not provided
        if roi is not None:
            # Apply y roi filter
            det_image = det_image[det_image[:,1] > roi["y"]["min"]]
            det_image = det_image[det_image[:,1] < roi["y"]["max"]]
            # Apply x roi filter
            det_image = det_image[det_image[:,0] > roi["x"]["min"]]
            det_image = det_image[det_image[:,0] < roi["x"]["max"]]
        # Apply spatial detector efficiency correction
        if efficiency_map is not None:
            eff, xe, ye = efficiency_map
            x_inds = np.digitize(det_image[:,0], xe)
            y_inds = np.digitize(det_image[:,1], ye)
            efficiencies = eff[x_inds, y_inds]
            # Get the inverse of the efficiency
            inv_eff = 1 / efficiencies[efficiencies > 0]
        else:
            inv_eff = np.ones(det_image.shape[0])
        # Calculate the number of counts
        n = np.sum(inv_eff)
        err = np.sqrt(len(inv_eff)) * n / len(inv_eff)
        # Normalize data to measurement duration
        if self._t is not None:
            n /= self._t
            err /= self._t
        # Subtract background before further normalization
        if background is not None:
            bkg_counts, bkg_err = background.counts(anode, roi=roi,
                efficiency_map=efficiency_map)
            n -= bkg_counts
            n = max(n, 0)
            err = np.sqrt(err**2 + bkg_err**2)
        # Normalize data to account for beam intensity, gas
        # pressure, etc.
        if self._target_density is not None:
            n /= self._target_density
            err /= self._target_density
        if self._intensity_upstream is not None:
            n /= self._intensity_upstream
            err /= self._intensity_upstream
        return n, err

    def spectrum(self,
        edges: np.ndarray,
        anode: PositionAnode = None,
        roi: dict = None,
        efficiency_map: Tuple[np.ndarray, np.ndarray, np.ndarray] = None,
        background: Spectrum = None,
        phem_calib: Tuple[float, float] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        TODO
        ----
        - Implement handling of the uncertainties of the efficiency map
        - Background subtraction is very primitive and uncertainties are
          not propagated correctly
        """
        # Get the xy values
        if anode is None:
            anode = self.anode
        if anode is None:
            raise ValueError("Anode object must be provided for processing.")
        det_image = anode.process(self._raw)
        # Define the roi as the full detector if not provided
        if roi is None:
            roi = {"x": {"min": 0, "max": 1}, "y": {"min": 0, "max": 1}}
        # Apply y roi filter
        det_image = det_image[det_image[:,1] > roi["y"]["min"]]
        det_image = det_image[det_image[:,1] < roi["y"]["max"]]
        # Apply spatial detector efficiency correction
        if efficiency_map is not None:
            eff, xe, ye = efficiency_map
            x_inds = np.digitize(det_image[:,0], xe)
            y_inds = np.digitize(det_image[:,1], ye)
            efficiencies = eff[x_inds, y_inds]
        else:
            efficiencies = np.ones(det_image.shape[0])
        # Don't need y values anymore
        det_image = det_image[:,0].flatten()
        # Convert x values to wavelengths
        if phem_calib is not None:
            a, b = phem_calib
            det_image = a * det_image + b
            # Adjust x roi filter to wavelength binning
            wl_min = edges[edges < (a * roi["x"]["min"] + b)][-1]
            wl_max = edges[edges > (a * roi["x"]["max"] + b)][0]
            xroi = (det_image >= wl_min) & (det_image <= wl_max)
        else:
            # Adjust x roi filter to detector binning
            x_min = edges[edges < roi["x"]["min"]][-1]
            x_max = edges[edges > roi["x"]["max"]][0]
            xroi = (det_image >= x_min) & (det_image <= x_max)
        # Apply x roi filter
        det_image = det_image[xroi]
        efficiencies = efficiencies[xroi]
        # Calculate the sum of weights for each bin, i.e. the weighted spectrum
        spectrum = np.histogram(det_image, bins=edges, weights=efficiencies)[0]
        # Histogram data without the per event efficiencies
        errors = np.sqrt(np.histogram(det_image, bins=edges)[0])
        # Normalize data to measurement duration per step
        if self._t is not None:
            spectrum /= self._t
            errors /= self._t
        # Subtract background before further normalization
        if background is not None:
            bkg_spec, bkg_err = background.spectrum(
                edges, anode, roi=roi, efficiency_map=efficiency_map,
                phem_calib=phem_calib
            )
            spectrum -= bkg_spec
            spectrum[spectrum < 0] = 0
            errors = np.sqrt(errors**2 + bkg_err**2)
        # Normalize data to account for beam intensity, gas 
        # pressure, etc.
        if self._target_density is not None:
            spectrum /= self._target_density
            errors /= self._target_density
        if self._intensity_upstream is not None:
            spectrum /= self._intensity_upstream
            errors /= self._intensity_upstream
        # Return the spectrum and uncertainties
        return spectrum, errors


class EnergyScan:
    """Scan over exciting-photon energies.

    Parameters
    ----------
    data_files: Sequence[str]
        List of data files to be processed.
    anode: PositionAnode
        Anode object from `agepy.spec.photons`.
    energies: str, optional
        Path to the exciting-photon energies in the data files. If None,
        the keys are used as the exciting-photon energies.
    raw: str, optional
        Path to the raw data in the data files. Default:
        "dld_rd#raw/0".
    time_per_step: int, optional
        Time per step in the scan. Default: None.
    target_density: str, optional
        Path to the target density in the data files. Default: None.
    intensity_downstream: str, optional
        Path to the downstream intensity in the data files. Default:
        None.
    intensity_upstream: str, optional
        Path to the upstream intensity in the data files. Default:
        None.

    Attributes
    ----------
    anode: PositionAnode
        Anode object from `agepy.spec.photons`.
    spectra: np.ndarray
        Array of the loaded Spectrum objects.
    energies: np.ndarray
        Array of the exciting-photon energies.

    TODO
    ----
    - Very minimal implementation, needs to be expanded

    """

    def __init__(self,
        data_files: Sequence[str],
        anode: PositionAnode,
        energies: str = None,
        raw: str = "dld_rd#raw/0",
        time_per_step: int = None,
        target_density: str = None,
        intensity_downstream: str = None,
        intensity_upstream: str = None,
    ) -> None:
        self.anode = anode  
        self.spectra = []
        self.energies = []
        if isinstance(data_files, str):
            data_files = [data_files]
        for f in data_files:
            spec, E = Spectrum.from_scan(
                f, scan_var=energies, raw=raw, time_per_step=time_per_step,
                anode=None, target_density=target_density,
                intensity_downstream=intensity_downstream,
                intensity_upstream=intensity_upstream
            )
            self.spectra.extend(spec)
            self.energies.extend(E)
        # Convert to numpy arrays
        self.energies = np.array(self.energies)
        self.spectra = np.array(self.spectra)
        # Sort the spectra by energy
        energy_sort = np.argsort(self.energies)
        self.energies = self.energies[energy_sort]
        self.spectra = self.spectra[energy_sort]

    def counts(self,
        roi: dict = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get the photon-excitation energy spectrum.

        Parameters
        ----------
        roi: dict
            Region of interest for the detector. If not provided, the
            full detector is used.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            The number of counts (normalized), the respective
            statistical uncertainties, and the exciting-photon energies.

        """
        vectorized_counts = np.vectorize(
            lambda spec: spec.counts(self.anode, roi=roi)
        )
        n, err = vectorized_counts(self.spectra)
        return n, err, self.energies

